from typing import List, Dict, Set, Optional, Iterable, Union, Tuple, ClassVar, cast
import re
from ..builtins import Char
from ..common import Input, dump_json
from ..jsonld.base import (
        VALUE, TYPE, LANGUAGE,
        ID, LIST, GRAPH,
        CONTEXT, VOCAB, BASE,
        PREFIX, PREFIX_DELIMS)
from ..rdfterms import RDF_TYPE, XSD, XSD_DOUBLE, XSD_INTEGER


ANNOTATION = '@annotation' # TODO: move to base (also from serializer)

XSD_DECIMAL: str = f'{XSD}decimal'

AT_KEYWORDS = {PREFIX, BASE}

RQ_PREFIX = 'prefix'
RQ_BASE = 'base'
RQ_GRAPH = 'graph'

RQ_KEYWORDS = {RQ_PREFIX, RQ_BASE, RQ_GRAPH}

ESC_CHARS = {
    't': '\t',
    'b': '\b',
    'n': '\n',
    'r': '\r',
    'f': '\f',
    '"': '"',
    "'": "'",
    '\\': '\\'
}

RESERVED_CHARS = {'~', '.', '-', '!',
                  '$', '&', "'", '(',
                  ')', '*', '+', ',',
                  ';', '=', '/', '?',
                  '#', '@', '%', '_'}

NUMBER_LEAD_CHARS = re.compile(r'[+-.0-9]')

TURTLE_INT_CHARS = re.compile(r'[+-.0-9]')

LITERAL_QUOTE_CHARS = {'"', "'"}

SYMBOL = '@symbol'
EOF = ''

StateResult = Tuple['ParserState', object]


class NotationError(Exception):
    pass


class ParserError(Exception):

    error: NotationError
    lno: int
    cno: int

    def __init__(self, error: NotationError, lno: int, cno: int):
        super().__init__(str(error))
        self.error = error
        self.lno = lno
        self.cno = cno

    def __str__(self) -> str:
        return (f'Notation error at line {self.lno}, column {self.cno}:'
                f' {self.error}')


class ParserState:

    parent: 'ParserState'
    context: Dict[str, object]

    def consume(self, c: str, prev_value) -> StateResult:
        raise NotImplementedError


class BaseParserState(ParserState):

    def __init__(self, parent: Optional[ParserState]):
        super().__init__()
        self.parent = parent if parent is not None else ParserState()
        self.context = self.parent.context if isinstance(self.parent, BaseParserState) else {}
        self.init()

    def init(self):
        pass

    def symbol(self, value: Dict[str, str]) -> str:
        if SYMBOL in value:
            sym = value[SYMBOL]
            if sym in self.context:
                sym = cast(str, self.context[VOCAB]) + sym
            return sym
        return value[ID]


class ConsumeWs(BaseParserState):

    MATCH: ClassVar[re.Pattern] = re.compile(r'\s')

    def accept(self, c: Char) -> bool:
        return self.MATCH.match(c) is not None

    def consume(self, c: str, prev_value) -> StateResult:
        if self.accept(c):
            return self, None
        else:
            return self.parent.consume(c, prev_value)


class ConsumeComment(BaseParserState):

    def consume(self, c: str, prev_value) -> StateResult:
        if c == '\n':
            return self.parent, None
        else:
            return self, None


class ReadTerm(BaseParserState):

    ESCAPE_CHAR: ClassVar[str] = '\\'

    collected: List[Char]
    escape_chars: Dict[Char, Char]
    escape_next: bool
    unicode_chars: List[Char]
    unicode_escapes_left: int

    def __init__(self, parent: Optional[ParserState]):
        super().__init__(parent)
        self.collected = []
        self.escape_next = False
        self.unicode_chars = []
        self.unicode_escapes_left = 0

    def collect(self, c: Char):
        self.collected.append(c)

    def pop(self) -> str:
        value = ''.join(self.collected)
        self.collected = []
        return value

    def handle_escape(self, c: Char) -> bool:
        if self.unicode_escapes_left:
            self.unicode_chars.append(c)
            if self.unicode_escapes_left == 1:
                hex_seq = ''.join(self.unicode_chars)
                try:
                    c = chr(int(hex_seq, 16))
                except ValueError:
                    raise NotationError(f'Invalid unicode escape: {hex_seq}')
                self.unicode_chars = []
                self.unicode_escapes_left = 0
            else:
                self.unicode_escapes_left -= 1
                return True

        elif self.escape_next:
            if c == 'u':
                self.unicode_escapes_left = 4
                return True
            elif c == 'U':
                self.unicode_escapes_left = 8
                return True
            elif c in self.escape_chars:
                c = self.escape_chars[c]
            else:
                raise NotationError(f'Invalid escape char: {c!r}')

        if self.escape_next:
            self.escape_next = False
            self.collect(c)
            return True

        if c == self.ESCAPE_CHAR:
            self.escape_next = True
            return True

        return False

    def backtrack(self, prev_c: str, c: str, value) -> StateResult:
        state, value = self.parent.consume(prev_c, value)
        return state.consume(c, value)


class ReadIRI(ReadTerm):

    MATCH: ClassVar[re.Pattern] = re.compile(r'\S') # TODO: specify better

    def init(self):
        self.escape_chars = {}

    def accept(self, c: Char) -> bool:
        return self.MATCH.match(c) is not None

    def consume(self, c: str, prev_value) -> StateResult:
        if c == '>':
            value = self.pop()
            return self.parent, {ID: value}
        elif self.handle_escape(c):
            return self, None
        else:
            if not self.accept(c):
                raise NotationError(f'Invalid URI character: {c!r}')
            self.collect(c)
            return self, None


class ReadSymbol(ReadTerm):

    MATCH: ClassVar[re.Pattern] = re.compile(r"[^\]\[{}^<>\"\s~!$&'()*,;=/?#]")

    just_escaped: bool

    def init(self):
        self.escape_chars = {c: c for c in RESERVED_CHARS}
        self.just_escaped = False

    def accept(self, c: Char) -> bool:
        if self.MATCH.match(c) is None:
            return False
        if c == ':' and len(self.collected) > 1 and (self.collected[0] == '_' and
                                                     self.collected[1] == ':'):
            return False
        return True

    def consume(self, c: str, prev_value) -> StateResult:
        if len(self.collected) == 0 and c == '<':
            return ReadIRI(self.parent), None
        elif len(self.collected) == 0 and NUMBER_LEAD_CHARS.match(c) is not None:
            return ReadNumber(self.parent).consume(c, None)
        elif self.handle_escape(c):
            self.just_escaped = True
            return self, None

        just_escaped = self.just_escaped
        self.just_escaped = False

        if self.accept(c):
            self.collect(c)
            return self, None

        value: Union[str, bool, Dict] = self.pop()
        assert isinstance(value, str)

        last_dot = False
        if not just_escaped and value.endswith('.'):
            value = value[:-1]
            last_dot = True

        if value in {'true', 'false'}:
            value = cast(bool, value == 'true')
        elif value == 'a':
            value = TYPE
        elif value not in AT_KEYWORDS:
            lowered = value.lower()
            if lowered in RQ_KEYWORDS:
                value = lowered
            else:
                if value != '':
                    if ':' not in value:
                        raise NotationError(f'Expected PNname, got {value!r}')
                    elif value[0] == ':':
                        value = value[1:]
                value = cast(Dict, {SYMBOL: value})

        assert isinstance(value, object)

        if last_dot:
            return self.backtrack('.', c, value)

        return self.parent.consume(c, value)


class ReadNumber(ReadTerm):

    EXP = {'E', 'e'}

    whole: Optional[str]
    dot: Char
    exp: bool

    def init(self):
        self.whole = None
        self.dot = ''
        self.exp = False

    def consume(self, c: str, prev_value) -> StateResult:
        exp = c in self.EXP
        if exp:
            self.exp = True
        if self.whole is None and c == '.':
            self.whole = self.pop()
            self.dot = c
            return self, None
        elif self.whole is None and exp:
            self.whole = self.pop()
            self.collect(c)
            return self, None
        elif c.isdecimal() or (self.whole is None and
                                len(self.collected) == 0 and
                                NUMBER_LEAD_CHARS.match(c) is not None) or (
                            self.whole is not None and c in self.EXP) or (
                                len(self.collected) > 0
                                and self.collected[-1] in self.EXP
                                and NUMBER_LEAD_CHARS.match(c) is not None):
            self.collect(c)
            return self, None
        else:
            number: object
            if self.whole is not None and len(self.collected) == 0:
                if self.whole == '':
                    return self.parent.consume(c, prev_value)
                number = int(self.whole)
                return self.backtrack('.', c, number)

            try:
                number = self.to_number()
            except ValueError as e:
                raise NotationError(f'Invalid number character, got {e}')

            return self.parent.consume(c, number)

    def to_number(self) -> Union[int, float, Dict]:
        value = self.pop()
        if self.whole:
            value = f'{self.whole}{self.dot}{value}'
            number: float = float(value)
            if number.is_integer():
                return {VALUE: value, TYPE: XSD_DOUBLE if self.exp else XSD_DECIMAL}
            return number
        else:
            if len(value) > 1 and TURTLE_INT_CHARS.match(value[0]) is not None:
                return {VALUE: value, TYPE: XSD_INTEGER}
            return int(value)


class ReadLiteral(ReadTerm):

    value: Optional[str]
    quotechar: str
    multiline: int
    prev_dt_start: int

    def __init__(self, parent: ParserState, quotechar: str):
        super().__init__(parent)
        self.quotechar = quotechar
        self.escape_chars = ESC_CHARS

    def init(self):
        self.prev_dt_start = 0
        self.value = None
        self.multiline = 0

    def consume(self, c: str, prev_value) -> StateResult:
        if self.value == '' and c == self.quotechar:
            self.multiline = 1
            self.value = None
            return self, None
        elif self.value is not None:
            if self.prev_dt_start:
                self.no_after_literal('Datatype', prev_value)
                if c == '^':
                    assert self.prev_dt_start == 1
                    self.prev_dt_start = 2
                    return self, None
                else:
                    assert self.prev_dt_start == 2
                    self.prev_dt_start = 0
                    return ReadSymbol(self).consume(c, None)

            if c == '^':
                self.no_after_literal('Datatype', prev_value)
                self.prev_dt_start = 1
                return self, None

            if c == '@':
                self.no_after_literal('Language', prev_value)
                state = ReadLanguage(self)
                return ReadLanguage(self), None

            value = {VALUE: self.value}
            if prev_value:
                if isinstance(prev_value, Dict) and LANGUAGE in prev_value:
                    value.update(prev_value)
                else:
                    assert isinstance(prev_value, Dict)
                    value[TYPE] = self.symbol(prev_value)

            return self.parent.consume(c, value)

        if self.handle_escape(c):
            return self, None

        if c == self.quotechar:
            if self.multiline == 0 or self.multiline == 3:
                self.multiline = 0
                self.value = self.pop()
            elif self.multiline > 0:
                self.multiline += 1
            return self, None

        if self.multiline > 1:
            for i in range(self.multiline - 1):
                self.collect(self.quotechar)
            self.multiline = 1

        self.collect(c)
        return self, None

    def no_after_literal(self, kind, prev_value):
        if prev_value is not None:
            raise NotationError(f'{kind} not allowed after {prev_value!r}')


class ReadLanguage(ReadTerm):

    MATCH: ClassVar[re.Pattern] = re.compile(r'[A-Za-z0-9-]') # TODO: only allow numbers in subtag

    def accept(self, c: Char) -> bool:
        return self.MATCH.match(c) is not None

    def consume(self, c: str, prev_value) -> StateResult:
        if self.accept(c):
            self.collect(c)
            return self, None
        else:
            value = self.pop()
            self.collected = []
            return self.parent.consume(c, {LANGUAGE: value})


class ReadCompound(BaseParserState):

    ws: ConsumeWs
    comment: ConsumeComment

    def __init__(self, parent: Optional[ParserState]):
        super().__init__(parent)
        self.ws = ConsumeWs(self)
        self.comment = ConsumeComment(self)

    def read_space(self, c: Char) -> Optional[StateResult]:
        if self.ws.accept(c):
            return self.ws, None
        elif c == '#':
            return self.comment, None
        else:
            return None

    def node_with_id(self, value: Dict) -> Dict:
        if SYMBOL in value:
            node_id: str = value[SYMBOL]
            if ':' not in node_id and VOCAB in self.context:
                node_id = cast(str, self.context[VOCAB]) + node_id
            value = {ID: node_id}
        return value

    def compact_value(self, value: object) -> object:
        if isinstance(value, Dict):
            if VALUE in value:
                if len(value) == 1:
                    return value[VALUE]
            else:
                return self.node_with_id(value)
        return value


class ReadDecl(ReadCompound):

    final_dot: bool
    completed: bool

    def __init__(self, parent: 'ReadNodes', final_dot: bool):
        super().__init__(parent)
        self.final_dot = final_dot
        self.completed = False

    def consume(self, c: str, prev_value) -> StateResult:
        if isinstance(prev_value, Dict):
            if not self.more_parts(prev_value):
                self.completed = True
                if not self.final_dot:
                    self.declare()
                    return self.parent.consume(c, None)

        readspace = self.read_space(c)
        if readspace:
            return readspace

        if c == '.':
            self.declare()
            return self.parent, None
        elif self.completed and self.final_dot:
            raise NotationError(f'Expected a final dot')

        return ReadSymbol(self).consume(c, None)

    def more_parts(self, value: Dict) -> bool:
        raise NotImplementedError

    def declare(self):
        raise NotImplementedError


class ReadPrefix(ReadDecl):

    pfx: Optional[str]
    ns: Optional[str]

    def init(self):
        self.pfx = None
        self.ns = None

    def more_parts(self, value: Dict) -> bool:
        if self.pfx is None:
            pfx = cast(str, value[SYMBOL])
            if pfx != '':
                if pfx.endswith(':'):
                    pfx = pfx[:-1]
                else:
                    raise NotationError(f'Invalid prefix {pfx!r}')
            self.pfx = pfx
            return True

        if self.ns is None:
            self.ns = value[ID]

        return False

    def declare(self):
        ns: Union[str, Dict[str, object]] = self.ns
        if self.pfx != '' and self.ns != '' and not self.ns[-1] in PREFIX_DELIMS:
            ns = {ID: self.ns, PREFIX: True}
        key: str = self.pfx if self.pfx is not None and self.pfx != '' else VOCAB
        self.parent.context[key] = ns


class ReadBase(ReadDecl):

    base: Optional[str]

    def init(self):
        self.base = None

    def more_parts(self, value: Dict) -> bool:
        if self.base is None:
            self.base = value[ID]
        return False

    def declare(self):
        self.parent.context[BASE] = self.base


class ReadNode(ReadCompound):

    node: Optional[Dict]
    p: Optional[str]
    last_value: Optional[object]
    open_brace: bool = False

    def fill_node(self, value):
        if self.p is None:
            if value == TYPE:
                self.p = TYPE
            else:
                if not isinstance(value, Dict):
                    raise NotationError(f'Unexpected predicate: {value!r}')
                self.p = self.symbol(value)
        elif self.last_value is None:
            if self.p == TYPE:
                assert isinstance(value, Dict)
                value = cast(str, self.symbol(value))

            value = self.compact_value(value)

            given: Optional[object] = self.node.get(self.p)
            if given is not None:
                values: List = given if isinstance(given, List) else [given]
                values.append(value)
                self.node[self.p] = values
            else:
                self.node[self.p] = value
            self.last_value = value
        elif isinstance(value, Dict) and ANNOTATION in value:
            last_value = self.last_value
            if self.p == TYPE:
                last_value = {TYPE: last_value}
            elif not isinstance(last_value, Dict):
                last_value = {VALUE: last_value}
            last_value[ANNOTATION] = value[ANNOTATION]
            if isinstance(self.node[self.p], List):
                l: List = self.node[self.p]
                l[-1] = last_value
            else:
                self.node[self.p] = last_value
        else:
            raise NotationError(f'Unexpected: {value!r}')

    def consume_node_char(self, c: str) -> StateResult:
        readspace = self.read_space(c)
        if readspace:
            return readspace

        if c == '{':
            self.open_brace = True
            return self, None
        elif c == '|':
            assert self.open_brace
            self.open_brace = False
            return ReadAnnotation(self), None
        elif c == '[':
            return ReadBNode(self), None
        elif c == '(':
            return ReadCollection(self), None
        elif c == ';':
            self.p = None
            self.last_value = None
            return self, None
        elif c == ',':
            self.last_value = None
            return self, None
        elif c in LITERAL_QUOTE_CHARS:
            return ReadLiteral(self, c), None
        else:
            return ReadSymbol(self).consume(c, None)


class ReadBNode(ReadNode):

    def init(self):
        self.reset()

    def reset(self):
        self.node = {}
        self.p = None
        self.last_value = None

    def consume(self, c: str, prev_value) -> StateResult:
        if prev_value is not None:
            self.fill_node(prev_value)

        if c == EOF:
            raise NotationError(f'Unexpected {c!r} in bnode.')
        elif c == ']':
            return self.parent, self.node
        else:
            return self.consume_node_char(c)


class ReadAnnotation(ReadBNode): # TODO: Factor out ReadNodeBase

    end_started: bool = False

    def consume(self, c: str, prev_value) -> StateResult:
        if prev_value is not None:
            self.fill_node(prev_value)

        if c == EOF:
            raise NotationError(f'Unexpected {c!r} in annotation.')
        elif not self.open_brace and c == '|':
            self.end_started = True
            return self, None
        elif c == '}':
            assert self.end_started
            self.end_started = False
            return self.parent, {ANNOTATION: self.node}
        else:
            return self.consume_node_char(c)


class ReadCollection(ReadCompound):

    nodes: List[object]

    def init(self):
        self.reset()

    def reset(self):
        self.nodes = []

    def consume(self, c: str, prev_value) -> StateResult:
        if prev_value is not None:
            self.nodes.append(self.compact_value(prev_value))

        readspace = self.read_space(c)
        if readspace:
            return readspace
        elif c == EOF:
            raise NotationError(f'Unexpected EOF in collection.')
        elif c == '[':
            return ReadBNode(self), None
        elif c == '(':
            return ReadCollection(self), None
        elif c == ')':
            return self.parent, {LIST: self.nodes}
        elif c in LITERAL_QUOTE_CHARS:
            return ReadLiteral(self, c), None
        else:
            return ReadSymbol(self).consume(c, None)


class ReadNodes(ReadNode):

    nodes: List[Dict]
    expect_graph: bool

    def init(self):
        self.nodes = []
        self.reset()

    def reset(self):
        self.node = None
        self.p = None
        self.last_value = None
        self.expect_graph = False

    def consume(self, c: str, prev_value) -> StateResult:
        if prev_value is not None:
            if isinstance(prev_value, str):
                final_dot = False
                if prev_value in AT_KEYWORDS:
                    prev_value = prev_value[1:]
                    final_dot = True
                if prev_value == RQ_PREFIX:
                    return ReadPrefix(self, final_dot).consume(c, None)
                elif prev_value == RQ_BASE:
                    return ReadBase(self, final_dot).consume(c, None)
                elif prev_value == RQ_GRAPH:
                    self.expect_graph = True
                    return self, None

            if self.node is None:
                assert isinstance(prev_value, Dict)
                self.node = self.node_with_id(prev_value)
            else:
                if self.p is None and self.expect_graph and self.node is not None:
                    raise NotationError('Expected graph notation to follow, '
                                        f'got {prev_value!r}')
                self.fill_node(prev_value)

        if c == EOF:
            result = {CONTEXT: self.context, GRAPH: self.nodes}
            return self.parent, result
        elif c == '.' and (self.p is None or self.last_value is not None):
            self.next_node()
            return self, None
        else:
            if self.open_brace:
                if c != '|':
                    self.open_brace = False
                    self.expect_graph = False
                    state = ReadGraph(self)
                    return state.consume(c, prev_value)
            return self.consume_node_char(c)

    def next_node(self):
        if self.node is None or (self.p is None and (
                                    (GRAPH not in self.node) and
                                    (ID in self.node and len(self.node) == 1))):
            raise NotationError(f'Incomplete triple for node: {self.node}')
        self.nodes.append(self.node)
        self.reset()


class ReadGraph(ReadNodes):

    def consume(self, c: str, prev_value) -> StateResult:
        if isinstance(prev_value, str) and prev_value != TYPE:
            raise NotationError(f'Directive not allowed in graph: {prev_value!r}')

        readnodes = cast(ReadNodes, self.parent)

        if self.expect_graph or self.open_brace and c != '|':
            raise NotationError('Nested graphs are not allowed in TriG')

        if c == '}':
            if self.node is not None:
                if self.p is not None and prev_value is not None:
                    self.fill_node(prev_value)
                self.next_node()
            if readnodes.node is None:
                readnodes.nodes += self.nodes
            else:
                readnodes.node[GRAPH] = self.nodes
                readnodes.next_node()
            return readnodes, None
        else:
            return super().consume(c, prev_value)


def parse(inp: Input) -> object:
    state: ParserState = ReadNodes(None)
    value: object = None

    lno = 1
    cno = 1
    for c in cast(Iterable[Char], inp.characters()):
        if c == '\n':
            lno += 1
            cno = 0

        next_state: ParserState
        try:
            next_state, value = state.consume(c, value)
        except NotationError as e:
            raise ParserError(e, lno, cno)

        cno += 1

        assert next_state is not None
        state = next_state

    endstate, result = state.consume(EOF, value)

    return result


if __name__ == '__main__':
    import sys

    inp = Input(sys.argv[1]) if len(sys.argv) > 1 else Input()
    try:
        result = parse(inp)
    except ParserError as e:
        print(e, file=sys.stderr)
    else:
        print(dump_json(result, pretty=True))



from typing import List, Optional, Iterable, Union, cast
from ..builtins import Char
from ..common import Input, dump_json
from ..jsonld.rdf import (
        RdfDataset, RdfGraph, RdfTriple, RdfLiteral, RdfObject, to_jsonld)

# TODO: Rewrite more based on trig.parser and its nested state handling
# Should basically become a ReadCompound subclass with restrictions on
# ReadSymbol allowing only BNodes...
from ..trig.parser import ESC_CHARS, ReadTerm


READ_STMT: int = 0
READ_IRI: int = 1
READ_BNODE_ID: int = 2
READ_STRING: int = 3
READ_LITERAL_END: int = 4
READ_DATATYPE_START: int = 5
READ_DATATYPE_NEXT: int = 6
READ_DATATYPE_IRI: int = 7
READ_LANGUAGE: int = 8
READ_LITERAL_FINISH: int = 9
READ_COMMENT: int = 10
#ESCAPE_NEXT: int = 11

#ESCAPE_CHAR: str = '\\'

READ_ESCAPES = ReadTerm(None)


def load(dataset: RdfDataset, inp: Input):
    state: Union[int, ReadTerm] = READ_STMT
    prev_state: Union[int, ReadTerm] = -1
    chars: List[str] = []
    literal: Optional[str] = None
    datatype: Optional[str] = None
    language: Optional[str] = None
    terms: List[RdfObject] = []

    for c in cast(Iterable[Char], inp.characters()):
        if READ_ESCAPES.handle_escape(c):
            if state is not READ_ESCAPES:
                READ_ESCAPES.escape_chars = ESC_CHARS if READ_STRING else {}
                prev_state = state
            state = READ_ESCAPES

        if state == READ_LITERAL_FINISH:
            assert literal is not None
            terms.append(RdfLiteral(literal, datatype, language))
            literal = datatype = language = None
            state = READ_STMT

        if state == READ_ESCAPES:
            if len(READ_ESCAPES.collected) == 1:
                c = READ_ESCAPES.pop()
                state = prev_state
            else:
                continue
        elif state == READ_STMT:
            if c.isspace():
                continue
            elif c == '<':
                state = READ_IRI
                continue
            elif c == '_':
                state = READ_BNODE_ID
            elif c == '"':
                state = READ_STRING
                continue
            elif c == '.':
                handle_statement(dataset, terms)
                terms = []
                continue
            elif c == '#':
                state = READ_COMMENT
                continue
        elif state == READ_IRI or state == READ_DATATYPE_IRI:
            if c == '>':
                s: str = ''.join(chars)
                if state == READ_IRI:
                    terms.append(s)
                    state = READ_STMT
                else:
                    datatype = s
                    state = READ_LITERAL_FINISH
                chars = []
                continue
        elif state == READ_BNODE_ID:
            if c.isspace():
                terms.append(''.join(chars))
                chars = []
                state = READ_STMT
                continue
        elif state == READ_STRING:
            if c == '"':
                literal = ''.join(chars)
                chars = []
                state = READ_LITERAL_END
                continue
        elif state == READ_LITERAL_END:
            if c == '@':
                state = READ_LANGUAGE
                continue
            elif c == '^':
                state = READ_DATATYPE_START
                continue
            else:
                state = READ_LITERAL_FINISH
                continue
        elif state == READ_LANGUAGE:
            if c.isspace():
                language = ''.join(chars)
                chars = []
                state = READ_LITERAL_FINISH
                continue
        elif state == READ_DATATYPE_START:
            if c == '^':
                state = READ_DATATYPE_NEXT
                continue
            else:
                raise Exception(f'Bad READ_DATATYPE_START char: {c}')
        elif state == READ_DATATYPE_NEXT:
            if c == '<':
                state = READ_DATATYPE_IRI
                continue
            else:
                raise Exception(f'Bad READ_DATATYPE_NEXT char: {c}')
        elif state == READ_COMMENT:
            if c == '\n':
                state = READ_STMT
            continue

        chars.append(c)

    if len(chars) != 0 or len(terms) != 0:
        raise Exception(f'Trailing data: chars={"".join(chars)!r}, terms={terms!r}')


def handle_statement(dataset: RdfDataset, terms: List):
    if len(terms) < 3 or len(terms) > 4:
        raise Exception(f'Invalid NQuads statement {str(terms)}')

    s: str = terms[0]
    p: str = terms[1]
    o: RdfObject = terms[2]
    g: Optional[str] = cast(str, terms[3]) if len(terms) == 4 else None

    graph: RdfGraph
    if g is None:
        graph = dataset.default_graph
        if graph is None:
            graph = dataset.default_graph = RdfGraph()
    else:
        if g not in dataset.named_graphs:
            dataset.named_graphs[g] = RdfGraph(g)
        graph = dataset.named_graphs[g]

    graph.add(RdfTriple(s, p, o))


def parse(inp: Input) -> object:
    dataset = RdfDataset()
    load(dataset, inp)
    return to_jsonld(dataset)


if __name__ == '__main__':
    inp = Input()
    result = parse(inp)
    print(dump_json(result, pretty=True))



import json
import sys
import traceback
from pathlib import Path

from ..common import Input, set_document_loader
from .base import JSONLD10, JSONLD11
from . import context
from .testbase import TestCase, TESTS_URL


realprint = print
def print(*args, **kwargs):
    realprint(*args, file=sys.stderr, **kwargs)


def _find_suite_file_dir(suitedir):
    while suitedir.parent.name and suitedir.name != 'tests':
        suitedir = suitedir.parent
    return str(suitedir)


def run_testsuite(suitefilepath: str) -> bool:
    suitefile = Path(suitefilepath)
    print(f'Running test suite: {suitefile}')
    suitedir = suitefile.parent
    with suitefile.open() as fp:
        testsuite = json.load(fp)

    suite_file_dir = _find_suite_file_dir(suitedir)
    set_document_loader(lambda url: Input(url.replace(TESTS_URL, suite_file_dir)))

    runs, oks, fails, errors = 0, 0, 0, 0

    for tc_data in testsuite['sequence']:
        tc = TestCase(str(suitedir), tc_data)
        if tc.options.get('specVersion') == JSONLD10:
            continue

        runs += 1
        print(f"Running TC {tc.testid} - {tc.name}:", end=' ')

        context.DEFAULT_PROCESSING_MODE = tc.options.get('processingMode', JSONLD11) # type: ignore

        try:
            out_data, expect_data = tc.run()
        except Exception as e:
            if tc.expected_error:
                if tc.expected_error.translate({ord(c): None for c in ' -@'}).lower() + 'error' == type(e).__name__.lower():
                    oks += 1
                    print(f'(Got expected error: {tc.expected_error}): OK')
                else:
                    fails += 1
                    report_failure(tc, expect_data)
                    print('>', end=' ')
                    traceback.print_exc(-3)
                    print()
            else:
                errors += 1
                print(f"ERROR in file '{tc.indoc_path}'")
                print('>', end=' ')
                traceback.print_exc(-3)
                print()
        else:
            out_shape = datarepr(out_data)
            expect_shape = datarepr(expect_data) if expect_data else out_shape

            if out_shape != expect_shape:
                fails += 1
                report_failure(tc, expect_shape)
                print(f'  Got: {out_shape}')
            else:
                oks += 1
                print('OK')

    print()
    print(f'Ran {runs} test cases. Passed: {oks}. Failed: {fails}. Errors: {errors}.')

    return not fails and not errors


def datarepr(data):
    if isinstance(data, str):
        return '\n'.join(sorted(data.splitlines()))
    return json.dumps(data, sort_keys=True, indent=2)


def report_failure(tc, expect_data):
    print('FAIL')
    print(f"  With base: '{tc.base_uri}'")
    print(f"  From file: '{tc.indoc_path}'")
    if tc.context_path:
        print(f"  Using context: '{tc.context_path}'")
    if tc.expectdoc_path:
        print(f"  Expecting: '{tc.expectdoc_path}'")
    if expect_data:
        print(f'  Expected: {expect_data}')
    if tc.expected_error:
        print(f'Expected error: {tc.expected_error}')


if __name__ == '__main__':
    args = sys.argv[1:]
    ok = True
    for arg in args:
        if not run_testsuite(arg):
            ok = False
    if not ok:
        sys.exit(1)



# TODO: Improve transpile to remove redundant redundancies:
# - Most cast calls (especially "upcasts" to JsonObject)
# - Some `if ...: if ....:` could be just `and`

from typing import Optional, Dict, List, Set, Union, cast
from ..common import warning
from .base import *
from .context import Context, Term, InvalidBaseDirectionError, InvalidNestValueError


# framing keywords
EMBED: str = '@embed'
EXPLICIT: str = '@explicit'
OMIT_DEFAULT: str = '@omit_default'
REQUIRES_ALL: str = '@requires_all'
FRAMING_KEYWORDS: Set[str] = {DEFAULT, EMBED, EXPLICIT, OMIT_DEFAULT, REQUIRES_ALL}


class InvalidReversePropertyMapError(JsonLdError): pass

class CollidingKeywordsError(JsonLdError): pass

class InvalidIdValueError(JsonLdError): pass

class InvalidTypeValueError(JsonLdError): pass

class InvalidIncludedValueError(JsonLdError): pass

class InvalidValueObjectError(JsonLdError): pass

class InvalidValueObjectValueError(JsonLdError): pass

class InvalidTypedValueError(JsonLdError): pass

class InvalidSetOrListObjectError(JsonLdError): pass

class InvalidLanguageTaggedValueError(JsonLdError): pass

class InvalidLanguageTaggedStringError(JsonLdError): pass

class InvalidIndexValueError(JsonLdError): pass

class InvalidReverseValueError(JsonLdError): pass

class InvalidReversePropertyValueError(JsonLdError): pass

class InvalidLanguageMapValueError(JsonLdError): pass


def expand(doc_data: JsonObject,
           base_iri: str,
           expand_context: str = None,
           ordered = False) -> List:
    ctx: Context = Context(base_iri)
    if expand_context is not None:
        ctx = ctx.get_context(expand_context, expand_context)
    result: Optional[JsonObject] = expansion(ctx, None, doc_data, base_iri,
                                             False, ordered)
    if result is None:
        return []
    if isinstance(result, Dict) and GRAPH in result and len(result) == 1:
        result = result[GRAPH]
    return as_list(result)


def expansion(active_context: Context,
        active_property: Optional[str],
        element: Optional[JsonObject],
        base_url: str,
        frame_expansion = False,
        ordered = False,
        from_map = False,
        # additional flags:
        warn_on_keywordlike_terms = False,
        warn_on_empty_keys = False,
        warn_on_bnode_properties = False,
        ) -> Optional[JsonObject]:
    # 1)
    if element is None:
        return None

    # 2)
    if active_property == '@default':
        frame_expansion = False

    property_scoped_context: Optional[Context] = None

    # 3)
    if active_property in active_context.terms and \
            active_context.terms[active_property].has_local_context:
        property_scoped_context = active_context.terms[active_property].get_local_context(active_context)

    # 4)
    if not isinstance(element, (Dict, List)):
        # 4.1)
        if active_property is None or active_property == GRAPH:
            return None
        # 4.2)
        if property_scoped_context is not None:
            active_context = property_scoped_context
        # 4.3)
        return value_expansion(active_context, active_property, element)

    # 5)
    if isinstance(element, List):
        # 5.1)
        results: List = []
        # 5.2)
        for item in element:
            # 5.2.1)
            expanded_item: Optional[JsonObject] = expansion(active_context, active_property, item, base_url, frame_expansion, from_map)
            # 5.2.2)
            active_term: Optional[Term] = active_context.terms.get(active_property)
            if active_term is not None and LIST in active_term.container and isinstance(expanded_item, List):
                expanded_item = {LIST: expanded_item}

            # 5.2.3)
            if isinstance(expanded_item, List):
                results += expanded_item
            elif expanded_item is not None:
                results.append(expanded_item)

        # 5.3)
        return results

    # 6)
    assert isinstance(element, Dict)

    # 7)
    # TODO: differs from spec (see spec problem 5f496df9)
    if not active_context._propagate and active_context.previous_context:
        # TODO: to make effectively final for closure generated by transpile
        active_context2: Context = active_context 
        if ((not from_map)
            and (not any(k in active_context2.terms and active_context2.terms[k].iri == VALUE for k in element))
            and (not (len(element) == 1 and any(active_context2.expand_iri(cast(str, k)) == ID for k in element)))):
            active_context = active_context.previous_context

    # 8)
    if property_scoped_context is not None:
        active_context = property_scoped_context

    # 9)
    if CONTEXT in element:
        active_context = active_context.get_context(element[CONTEXT], base_url)

    # 10)
    type_scoped_context: Context = active_context

    # 11)
    values: Optional[List[str]] = None
    keys: List[str] = list(element.keys())
    keys.sort()
    for key in keys:
        if type_scoped_context.expand_iri(key) == TYPE:
            # 11.1)
            values = list(as_list(element[key]))
            values.sort()
            # 11.2)
            for value in values:
                term: Optional[Term] = type_scoped_context.terms.get(value)
                if isinstance(value, str) and term and term.has_local_context:
                    active_context = term.get_local_context(type_scoped_context, False)

    # 12)
    result: Dict = {}
    nests: Dict = {}
    input_type: Optional[str] = values[-1] if values else None
    if input_type is not None and isinstance(input_type, str):
        # TODO: or expand immediately above?
        input_type = active_context.expand_iri(input_type)


    # 13) + 14)
    _expand_element(
            active_context, type_scoped_context, active_property, element, result, nests, input_type,
            base_url, frame_expansion, ordered, from_map,
            warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties)

    # 15)
    if VALUE in result:
        resultval: JsonObject = result[VALUE]
        # 15.1)
        if any(k not in VALUE_KEYWORDS for k in result.keys()) \
           or TYPE in result and (LANGUAGE in result or DIRECTION in result):
            raise InvalidValueObjectError
        # 15.2)
        if TYPE in result and result[TYPE] == JSON:
            pass # VALUE entry may contain any value
        # 15.3)
        elif resultval is None or resultval == []:
            return None
        # 15.4)
        elif not isinstance(resultval, str) and LANGUAGE in result:
            raise InvalidLanguageTaggedValueError
        # 15.5)
        elif TYPE in result and not (isinstance(result[TYPE], str) and is_iri(cast(str, result[TYPE]))):
            raise InvalidTypedValueError(str(result[TYPE]))

    # 16)
    elif TYPE in result and not isinstance(result[TYPE], List):
        result[TYPE] = [result[TYPE]]

    # 17)
    elif SET in result or LIST in result:
        # 17.1)
        if len(result) > 1:
            if len(result) > 2 or INDEX not in result:
                raise InvalidSetOrListObjectError(str(result))
        # 17.2)
        if SET in result:
            # TODO: spec says "set result to", but the remaining steps are only
            # applicable when result is still a Dict...
            #result =
            return result[SET]

    # 18)
    if len(result) == 1 and LANGUAGE in result:
        return None

    # 19)
    if active_property is None or active_property == GRAPH:
        # 19.1)
        # TODO: spec errata: step says "or contains only the entries @value or list@"
        # By dropping "only", TC 0046 passes.
        # Thus removed from paren after `or`: `len(result) == 1 and `
        if (isinstance(result, Dict) and len(result) == 0) or ((VALUE in result or LIST in result)):
            return None
        # 19.2)
        elif not frame_expansion and len(result) == 1 and ID in result:
            return None

    # 20)
    return result


def _expand_element(active_context: Context,
        type_scoped_context: Context,
        active_property: Optional[str],
        element: JsonOptMap,
        result: JsonOptMap,
        nests: JsonMap,
        input_type: Optional[str],
        base_url: str,
        frame_expansion = False,
        ordered = False,
        from_map = False,
        # additional flags:
        warn_on_keywordlike_terms = False,
        warn_on_empty_keys = False,
        warn_on_bnode_properties = False,
        ):
    # 13)
    keys: List[str] = list(element.keys())
    if ordered:
        keys.sort()
    for key in keys:
        value: Optional[JsonObject] = element[key]
        expanded_value: Optional[JsonObject] = None
        expanded_map: JsonOptMap
        expanded_list: JsonOptList
        reverse_map: JsonMap

        # 13.1)
        if key == CONTEXT:
            continue
        # 13.2)
        expanded_property: Optional[str] = active_context.expand_vocab_iri(key)
        # 13.3)
        if expanded_property is None or (':' not in expanded_property and expanded_property not in KEYWORDS):
            continue
        # 13.4)
        if expanded_property in KEYWORDS:
            # 13.4.1)
            if active_property == REVERSE:
                raise InvalidReversePropertyMapError
            # 13.4.2)
            if expanded_property in result and expanded_property not in {INCLUDED, TYPE}:
                pass # FIXME: expansion TC in06 fails with this on ID! Due to nested recusion?
                # TODO: spec problem; just allow ID too?
                #raise CollidingKeywordsError(expanded_property)

            # 13.4.3)
            if expanded_property == ID:
                # 13.4.3.1)
                if not isinstance(value, str) or (frame_expansion and
                        (value == {} or isinstance(value, List))):
                    raise InvalidIdValueError
                # 13.4.3.2)
                if frame_expansion:
                    expanded_value = expanded_list = []
                    for v in expanded_list:
                        expanded_list.append(
                            active_context.expand_doc_relative_iri(v)
                            if isinstance(v, str) else v)
                else:
                    expanded_value = active_context.expand_doc_relative_iri(value)

            # 13.4.4)
            elif expanded_property == TYPE:
                # 13.4.4.1)
                if not isinstance(value, (str, List)) and not (
                        frame_expansion and (value == {} or
                            (isinstance(value, Dict) and
                             is_iri(cast(str, value.get(DEFAULT)))))): # type: ignore
                    raise InvalidTypeValueError
                # 13.4.4.2)
                if isinstance(value, Dict) and value == {}:
                    expanded_value = expanded_map = value
                # 13.4.4.3)
                elif isinstance(value, Dict) and DEFAULT in value:
                    expanded_map = {DEFAULT: type_scoped_context.expand_doc_relative_vocab_iri(cast(str, value[DEFAULT]))}
                # 13.4.4.4)
                else:
                    expanded_value = expanded_list = []
                    for v in as_list(cast(JsonObject, value)):
                        expanded_list.append(type_scoped_context.expand_doc_relative_vocab_iri(cast(str, v)))
                    # TODO: spec errata: be more explicit about collapsing singe-item list
                    if len(expanded_list) == 1:
                        expanded_value = expanded_list[0]
                # 13.4.4.5)
                if TYPE in result:
                    expanded_value = as_list(result[TYPE]) + as_list(expanded_value)
                # TODO: will result be set further below?
                #else:
                #    result[TYPE] = expanded_value

            # 13.4.5)
            elif expanded_property == GRAPH:
                expanded_value = as_list(expansion(active_context,
                    GRAPH, value,
                    base_url, frame_expansion, ordered, from_map,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties))

            # 13.4.6)
            elif expanded_property == INCLUDED:
                # 13.4.6.1)
                if active_context._processing_mode == JSONLD10:
                    continue
                # 13.4.6.2)
                expanded_value = expanded_list = as_list(expansion(active_context,
                    None, value,
                    base_url, frame_expansion, ordered, from_map,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties))
                # 13.4.6.3)
                if any(not isinstance(v, Dict) for v in expanded_list):
                    raise InvalidIncludedValueError
                # 13.4.6.4)
                if INCLUDED in result:
                    expanded_value = as_list(result[INCLUDED]) + expanded_list

            # 13.4.7)
            elif expanded_property == VALUE:
                # 13.4.7.1)
                if input_type == JSON:
                    if active_context._processing_mode == JSONLD10:
                        raise InvalidValueObjectValueError
                    expanded_value = value
                # 13.4.7.2)
                elif (not is_scalar(value) and value is not None) and not \
                    (frame_expansion and (value == {} or
                        (isinstance(value, List) and
                         all(is_scalar(v) for v in value)))):
                    raise InvalidValueObjectValueError(f'Unexpected value {value}')
                # 13.4.7.3)
                else:
                    expanded_value = cast(JsonObject, value)
                    if frame_expansion:
                        expanded_value = as_list(expanded_value)
                # 13.4.7.4)
                if expanded_value is None:
                    result[VALUE] = None
                    continue

            # 13.4.8)
            elif expanded_property == LANGUAGE:
                # 13.4.8.1)
                if not isinstance(value, str):
                    raise InvalidLanguageTaggedStringError
                # 13.4.8.2)
                expanded_value = value
                if not is_lang_tag(value):
                    warning(f'Language tag {value} is not well-formed')
                if frame_expansion:
                    expanded_value = as_list(expanded_value)

            # 13.4.9)
            elif expanded_property == DIRECTION:
                # 13.4.9.1)
                if active_context._processing_mode == JSONLD10:
                    continue
                # 13.4.9.2)
                if value not in DIRECTIONS or frame_expansion and (value != {} or isinstance(value, List) and not all(isinstance(v, str) for v in value)):
                    raise InvalidBaseDirectionError
                # 13.4.9.3)
                expanded_value = value
                if frame_expansion:
                    expanded_value = as_list(expanded_value)

            # 13.4.10)
            elif expanded_property == INDEX:
                # 13.4.10.1)
                if not isinstance(value, str):
                    raise InvalidIndexValueError
                # 13.4.10.2)
                expanded_value = value

            # 13.4.11)
            elif expanded_property == LIST:
                # 13.4.11.1)
                if active_property is None or active_property == GRAPH:
                    continue
                # 13.4.11.2)
                expanded_value = as_list(expansion(active_context,
                    active_property, value,
                    base_url, frame_expansion, ordered, from_map,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties))

            # 13.4.12)
            elif expanded_property == SET:
                expanded_value = expansion(active_context,
                        active_property, value,
                        base_url, frame_expansion, ordered, from_map,
                        warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties)

            # 13.4.13)
            elif expanded_property == REVERSE:
                # 13.4.13.1)
                if not isinstance(value, Dict):
                    raise InvalidReverseValueError
                # 13.4.13.2)
                # TODO:[5f4f785a] algorithm uses expanded_value regardless of
                # type, but declaring local, specifically typed variables are
                # simpler to transcribe. (We probably want to factor out the
                # steps instead though, to reduce and isolate scopes (and smoke
                # out complexity).)
                expanded_map = cast(JsonOptMap, expansion(active_context,
                    REVERSE, value,
                    base_url, frame_expansion, ordered, from_map,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties))

                # 13.4.13.3)
                if REVERSE in expanded_map:
                    # 13.4.13.3.1)
                    entry: JsonMap = cast(JsonMap, expanded_map[REVERSE])
                    for prop, item in entry.items():
                        add_value_as_list(result, prop, item)

                # 13.4.13.4)
                if any(k != REVERSE for k in expanded_map.keys()):
                    # 13.4.13.4.1)
                    reverse_map = cast(JsonMap, result.setdefault(REVERSE, {}))
                    # 13.4.13.4.2)
                    for prop, items in expanded_map.items():
                        # 13.4.13.4.2.1)
                        for item in cast(List[JsonMap], items):
                            # 13.4.13.4.2.1.1)
                            if isinstance(item, Dict) and (VALUE in item or LIST in item):
                                raise InvalidReversePropertyValueError
                            # 13.4.13.4.2.1.2)
                            add_value_as_list(reverse_map, prop, item)

                # 13.4.13.5)
                continue

            # 13.4.14)
            elif expanded_property == NEST:
                nests.setdefault(key, [])
                continue

            # 13.4.15)
            if frame_expansion and expanded_property in FRAMING_KEYWORDS:
                expanded_value = expansion(active_context,
                        active_property, value,
                        base_url, frame_expansion, ordered, from_map,
                        warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties)

            # 13.4.16)
            if not (expanded_value is None and expanded_property == VALUE and input_type != JSON):
                result[expanded_property] = expanded_value

            # 13.4.17)
            continue

        # 13.5)
        key_term: Optional[Term] = active_context.terms.get(key)
        container_mapping: List = key_term.container if key_term else []

        # 13.6)
        if key_term and key_term.type_mapping == JSON:
            expanded_value = {VALUE: value, TYPE: JSON}

        # 13.7)
        elif LANGUAGE in container_mapping and isinstance(value, Dict):
            # 13.7.1)
            expanded_value = expanded_list = []
            # 13.7.2)
            direction: Optional[str] = active_context.default_base_direction
            # 13.7.3)
            if key_term and key_term.direction is not None:
                direction = key_term.direction
            # 13.7.4)
            languages: List[str] = list(value.keys())
            if ordered:
                languages.sort()
            for lang in languages:
                # 13.7.4.1)
                langvalues: List[str] = as_list(value[lang])
                # 13.7.4.2)
                for item in langvalues:
                    # 13.7.4.2.1)
                    if item is None:
                        continue
                    # 13.7.4.2.2)
                    if not isinstance(item, str):
                        raise InvalidLanguageMapValueError
                    # 13.7.4.2.3)
                    o: Dict[str, str] = {VALUE: item, LANGUAGE: lang}
                    # TODO: spec errata: s/If item/If language/
                    if lang not in NULLS and not is_lang_tag(lang):
                        warning(f'Language tag {value} is not well-formed')
                    # 13.7.4.2.4)
                    if lang in NULLS or active_context.expand_iri(lang) == NONE:
                        del o[LANGUAGE]
                    # 13.7.4.2.5)
                    if direction not in NULLS: # TODO: marked from None
                        o[DIRECTION] = cast(str, direction)
                    # 13.7.4.2.6)
                    expanded_list.append(o)

        # 13.8)
        elif any(k in container_mapping for k in [INDEX, TYPE, ID]) and isinstance(value, Dict):
            # 13.8.1)
            expanded_value = expanded_list = []
            # 13.8.2)
            index_key: str = INDEX
            if key_term and key_term.index is not None:
                index_key = key_term.index

            # 13.8.3)
            indexes: List[str] = list(value.keys())
            if ordered:
                indexes.sort()
            for index in indexes:
                # 13.8.3.1)
                map_context: Context
                if (ID in container_mapping or TYPE in container_mapping) and \
                        active_context.previous_context:
                    map_context = active_context.previous_context
                else:
                    map_context = active_context

                # 13.8.3.2)
                index_term: Optional[Term] = active_context.terms.get(index)
                if TYPE in container_mapping and index_term and index_term.has_local_context:
                    map_context = index_term.get_local_context(map_context)
                # 13.8.3.3)
                else:
                    map_context = active_context

                # 13.8.3.4)
                expanded_index: str = cast(str, active_context.expand_vocab_iri(index))
                # 13.8.3.5)
                index_values: List[JsonMap] = as_list(value[index])
                # 13.8.3.6)
                index_values = cast(List[JsonMap], expansion(map_context,
                    key, index_values,
                    base_url, frame_expansion, ordered, True,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties))
                # 13.8.3.7)
                for item in index_values:
                    # 13.8.3.7.1)
                    if GRAPH in container_mapping and (not isinstance(item, Dict) or GRAPH not in item):
                        item = {GRAPH: as_list(item)}
                    # 13.8.3.7.2)
                    if INDEX in container_mapping and index_key != INDEX and expanded_index != NONE:
                        # 13.8.3.7.2.1)
                        reexpanded_index: JsonObject = value_expansion(
                                active_context, index_key, index)
                        # 13.8.3.7.2.2)
                        expanded_index_key: str = cast(str, active_context.expand_vocab_iri(index_key))
                        # 13.8.3.7.2.3)
                        index_property_values: List = [reexpanded_index]
                        if expanded_index_key in item:
                            index_property_values += as_list(item[expanded_index_key])
                        # 13.8.3.7.2.4)
                        item[expanded_index_key] = index_property_values
                        # 13.8.3.7.2.5)
                        if isinstance(item, Dict) and VALUE in item and len(item) > 1:
                            raise InvalidValueObjectError
                    # 13.8.3.7.3)
                    elif INDEX in container_mapping and INDEX not in item and expanded_index != NONE:
                        item[INDEX] = index
                    # 13.8.3.7.4)
                    elif ID in container_mapping and ID not in item and expanded_index != NONE:
                        # TODO: spec errata: different from step 13.8.3.4?
                        expanded_index = cast(str, active_context.expand_doc_relative_iri(index))
                        item[ID] = expanded_index
                    # 13.8.3.7.5)
                    elif TYPE in container_mapping and expanded_index != NONE:
                        types: List[str] = [expanded_index]
                        if TYPE in item:
                            types += as_list(item[TYPE])
                        item[TYPE] = types
                    # 13.8.3.7.6)
                    expanded_list.append(item)

        # 13.9)
        else:
            expanded_value = expansion(active_context,
                    key, value,
                    base_url, frame_expansion, ordered, from_map,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties)

        # 13.10)
        if expanded_value is None:
            continue

        # 13.11)
        if LIST in container_mapping:
            if not isinstance(expanded_value, Dict) or LIST not in expanded_value:
                expanded_value = {LIST: as_list(cast(JsonObject, expanded_value))}

        # 13.12)
        if GRAPH in container_mapping and ID not in container_mapping and INDEX not in container_mapping:
            new_expanded: List = []
            for ev in as_list(expanded_value):
                # 13.12.1)
                new_expanded.append({GRAPH: as_list(ev)})
            expanded_value = new_expanded

        # 13.13)
        if key_term and key_term.is_reverse_property:
            # 13.13.1)
            # 13.13.2)
            reverse_map = cast(JsonMap, result.setdefault(REVERSE, {}))
            # 13.13.3)
            if not isinstance(expanded_value, List):
                expanded_value = [expanded_value]
            # 13.13.4)
            for item in expanded_value:
                # 13.13.4.1)
                if isinstance(item, Dict) and (VALUE in item or LIST in item):
                    raise InvalidReversePropertyValueError
                # 13.13.4.2)
                rev_entry: List = cast(List, reverse_map.setdefault(expanded_property, []))
                # 13.13.4.3)
                add_value_as_list(reverse_map, expanded_property, item)

        # 13.14)
        else:
            add_value_as_list(result, expanded_property, expanded_value)

    # 14)
    nesting_keys: List[str] = list(nests.keys())
    if ordered:
        nesting_keys.sort()
    for nesting_key in nesting_keys:
        # 14.1)
        nested_values: List[JsonObject] = as_list(element[nesting_key]) if nesting_key in element else []
        # 14.2)
        for nested_value in nested_values:
            # 14.2.1)
            if not isinstance(nested_value, Dict) or any(active_context.expand_iri(cast(str, k)) == VALUE for k in nested_value.keys()):
                raise InvalidNestValueError
            # 14.2.2)
            # TODO:
            _expand_element(
                    active_context, type_scoped_context, active_property, nested_value, result, nests, input_type,
                    base_url, frame_expansion, ordered, from_map,
                    warn_on_keywordlike_terms, warn_on_empty_keys, warn_on_bnode_properties)


def value_expansion(active_context: Context, active_property: str, value: Scalar) -> JsonObject:
    # 1)
    active_term: Optional[Term] = active_context.terms.get(active_property)
    if active_term and active_term.type_mapping == ID and isinstance(value, str):
        return {ID: active_context.expand_doc_relative_iri(value)}

    # 2)
    if active_term and active_term.type_mapping == VOCAB and isinstance(value, str):
        return {ID: active_context.expand_doc_relative_vocab_iri(value)}

    # 3)
    result: Dict[str, Optional[JsonObject]] = {VALUE: value}

    # 4)
    if active_term and active_term.type_mapping not in {ID, VOCAB, NONE, None}:
        result[TYPE] = active_term.type_mapping
    # 5)
    elif isinstance(value, str):
        # 5.1)
        language: Optional[str] = active_context.default_language
        if active_term and active_term.language is not None:
            # TODO: see 5f6117d4
            language = None if active_term.language in NOTHING else active_term.language
        # 5.2)
        direction: Optional[str] = active_context.default_base_direction
        if active_term and active_term.direction is not None:
            # TODO: see previous about Term step 22 (this corresponds to step 23)
            direction = None if active_term.direction in NOTHING else active_term.direction
        # 5.3)
        if language is not None:
            result[LANGUAGE] = language
        # 5.4)
        if direction is not None:
            result[DIRECTION] = direction

    # 6)
    return result



RDF: str = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
RDF_TYPE: str = f'{RDF}type'
RDF_VALUE: str = f'{RDF}value'
RDF_LIST: str = f'{RDF}List'
RDF_FIRST: str = f'{RDF}first'
RDF_REST: str = f'{RDF}rest'
RDF_NIL: str = f'{RDF}nil'
RDF_DIRECTION: str = f'{RDF}direction'
RDF_LANGUAGE: str = f'{RDF}language'
RDF_JSON: str = f'{RDF}JSON'
RDF_LANGSTRING: str = f'{RDF}langString'

XSD: str = 'http://www.w3.org/2001/XMLSchema#'
XSD_BOOLEAN: str = f'{XSD}boolean'
XSD_DECIMAL: str = f'{XSD}decimal'
XSD_DOUBLE: str = f'{XSD}double'
XSD_INTEGER: str = f'{XSD}integer'
XSD_STRING: str = f'{XSD}string'

I18N: str = 'https://www.w3.org/ns/i18n#'



from typing import Optional
from ..common import Output
from ..jsonld.base import is_blank
from ..jsonld.rdf import RdfDataset, RdfGraph, RdfTriple, RdfLiteral
from ..rdfterms import XSD_STRING


def serialize(dataset: RdfDataset, out: Output):
    for graph in dataset:
        write_graph(graph, out)


def write_graph(graph: RdfGraph, out: Output):
    for triple in graph.triples:
        if triple.s is None or triple.p is None or triple.o is None:
            continue
        out.writeln(repr_quad(triple, graph.name))


def repr_quad(triple: RdfTriple, graph_name: Optional[str]) -> str:
    s: str = repr_term(triple.s)
    p: str = repr_term(triple.p)
    o: str = repr_term(triple.o)
    spo: str = f'{s} {p} {o}'
    quad: str = f'{spo} {repr_term(graph_name)}' if graph_name else spo
    return f'{quad} .'


def repr_term(t: object) -> str:
    if isinstance(t, str):
        if is_blank(t):
            return t
        else:
            return f'<{t}>'
    else:
        assert isinstance(t, RdfLiteral)
        v: str = t.value
        v = v.replace('\\', r'\\')
        v = v.replace('"', r'\"')
        v = f'"{v}"'
        if t.language is not None:
            return f'{v}@{t.language}'
        elif t.datatype is not None and t.datatype != XSD_STRING:
            dt: str = repr_term(t.datatype)
            return f'{v}^^{dt}'
        else:
            return v



from typing import Optional, Dict, List, Set, Tuple, Union, cast
from ..common import load_json, warning, resolve_iri
from .base import *


DEFAULT_PROCESSING_MODE: str = JSONLD11

MAX_REMOTE_CONTEXTS: int = 512


class ProcessingModeConflictError(JsonLdError): pass

class InvalidVersionValueError(JsonLdError): pass

class InvalidBaseIriError(JsonLdError): pass

class InvalidVocabMappingError(JsonLdError): pass

class InvalidDefaultLanguageError(JsonLdError): pass

class InvalidBaseDirectionError(JsonLdError): pass

class InvalidContextEntryError(JsonLdError): pass

class InvalidPropagateValueError(JsonLdError): pass

class InvalidImportValueError(JsonLdError): pass

class InvalidContextNullificationError(JsonLdError): pass

class InvalidLocalContextError(JsonLdError): pass

class LoadingDocumentFailedError(JsonLdError): pass

class ContextOverflowError(JsonLdError): pass

class InvalidRemoteContextError(JsonLdError): pass

class CyclicIriMappingError(JsonLdError): pass

class KeywordRedefinitionError(JsonLdError): pass

class InvalidTermDefinitionError(JsonLdError): pass

class InvalidScopedContextError(JsonLdError): pass

class InvalidProtectedValueError(JsonLdError): pass

class InvalidLanguageMappingError(JsonLdError): pass

class InvalidTypeMappingError(JsonLdError): pass

class InvalidReversePropertyError(JsonLdError): pass

class InvalidIriMappingError(JsonLdError): pass

class InvalidKeywordAliasError(JsonLdError): pass

class InvalidContainerMappingError(JsonLdError): pass

class InvalidNestValueError(JsonLdError): pass

class InvalidPrefixValueError(JsonLdError): pass

class ProtectedTermRedefinitionError(JsonLdError): pass


class Context:

    terms: Dict[Optional[str], 'Term']
    base_iri: str # IRI
    original_base_url: Optional[str] # IRI

    _inverse_context: Optional[Dict]

    vocabulary_mapping: Optional[str] # IRI
    default_language: Optional[str] # Language
    default_base_direction: Optional[str] # Direction

    _propagate: bool
    previous_context: Optional['Context']

    _processing_mode: str
    _version: Optional[float]

    #_keyword_aliases: Dict[str, List[str]]

    def __init__(self, base_iri: Optional[str], original_base_url: Optional[str] = None):
        self.initialize(base_iri, original_base_url)

    def initialize(self, base_iri: Optional[str], original_base_url: Optional[str] = None):
        self.terms = {}
        # TODO: spec problem; what if None?
        self.base_iri = "" if base_iri is None else base_iri
        if original_base_url is not None:
            self.original_base_url = original_base_url # TODO: resolve/check
        else:
            self.original_base_url = None # TODO: base_iri ?

        self.vocabulary_mapping = None
        self.default_language = None
        self.default_base_direction = None
        self._propagate = True
        self.previous_context = None
        self._processing_mode = DEFAULT_PROCESSING_MODE
        self._version = None
        self._inverse_context = None
        #self._keyword_aliases = {}

    def copy(self) -> 'Context':
        cloned: Context = Context(self.base_iri, self.original_base_url)
        cloned.terms = self.terms.copy()
        cloned.vocabulary_mapping = self.vocabulary_mapping
        cloned.default_language = self.default_language
        cloned.default_base_direction = self.default_base_direction
        cloned._processing_mode = self._processing_mode
        return cloned

    def get_context(self, context_data: object,
            base_url: str = None,
            remote_contexts: Set[str] = None,
            override_protected=False,
            validate_scoped=True) -> 'Context':
        if remote_contexts is None:
            remote_contexts = set()
        # 1)
        local_context: Context = self.copy()
        # 2)
        if isinstance(context_data, Dict):
            propagate: object = context_data.get(PROPAGATE)
            if isinstance(propagate, bool):
                local_context._propagate = propagate
        # 3)
        # TODO: spec problem 5f496df9:
        # always keep previous_context and use depending on propagate
        #if not local_context._propagate and \
        #        local_context.previous_context is None:
        if local_context.previous_context is None:
            local_context.previous_context = self

        local_context._read_context(context_data, base_url,
                remote_contexts, override_protected, validate_scoped)

        return local_context

    def _read_context(self,
            context_data: object,
            base_url: Optional[str],
            remote_contexts: Set[str],
            override_protected: bool,
            validate_scoped: bool):
        # 4)
        normalized_context_data: List[object]
        if isinstance(context_data, List):
            normalized_context_data = context_data
        else:
            normalized_context_data = [context_data]

        if base_url is None:
            base_url = self.base_iri

        # 5)
        for context in normalized_context_data:
            # 5.1)
            if context is None:
                # 5.1.1)
                if override_protected is False and \
                        any(term.is_protected for term in self.terms.values()):
                    raise InvalidContextNullificationError
                # 5.1.2)
                # TODO: change to non-mutating behaviour? (self vs. result)
                prev: Optional[Context] = self.copy() if self._propagate is False else None
                self.initialize(base_url, base_url)
                if prev is not None:
                    self.previous_context = prev
                # 5.1.3)
                continue

            # 5.2)
            if isinstance(context, str):
                self._read_context_link(context, base_url, remote_contexts,
                        override_protected, validate_scoped)
            # 5.4)
            elif isinstance(context, Dict):
                self._read_context_definition(context, base_url, remote_contexts,
                        override_protected, validate_scoped)
            # 5.3)
            else:
                raise InvalidLocalContextError

    def _read_context_link(self, href: str, base_url: str,
            remote_contexts: Set[str],
            override_protected: bool, validate_scoped: bool):
        # 5.2.1)
        try:
            href = resolve_iri(base_url, href)
        except:
            raise LoadingDocumentFailedError

        if not validate_scoped and href in remote_contexts:
            return

        if len(remote_contexts) > MAX_REMOTE_CONTEXTS:
            raise ContextOverflowError
        else:
            remote_contexts.add(href)

        context_document: object = self._load_document(href)
        # 5.2.5.2)
        if not isinstance(context_document, Dict) or CONTEXT not in context_document:
            raise InvalidRemoteContextError
        # 5.2.5.3)
        loaded: object = context_document[CONTEXT]

        # 5.2.6)
        self._read_context(loaded, href, set(remote_contexts), override_protected, validate_scoped)
        # NOTE: If context was previously dereferenced, processors MUST make
        # provisions for retaining the base URL of that context for this step
        # to enable the resolution of any relative context URLs that may be
        # encountered during processing.

        # 5.2.7) Continue with the next context

    def _load_document(self, href: str, profile: str = JSONLD_CONTEXT_RELATION, request_profile: str = JSONLD_CONTEXT_RELATION) -> object:
        # TODO:
        # 5.2.4) If context was previously dereferenced, then the processor MUST NOT do a further dereference, and context is set to the previously established internal representation:
            #set context document to the previously dereferenced document, and set loaded context to the value of the @context entry from the document in context document.
        # NOTE: Only the @context entry need be retained.
        #loaded: Union[Dict, List] = context_document[CONTEXT]

        # 5.2.5) Otherwise, set context document to the RemoteDocument obtained by dereferencing context using the LoadDocumentCallback, passing context for url, and http://www.w3.org/ns/json-ld#context for profile and for requestProfile.
            # 5.2.5.1) If context cannot be dereferenced, or the document from context document cannot be transformed into the internal representation:
                #a loading remote context failed error has been detected and processing is aborted.
        ...
        #return remote_document.json
        return load_json(href)

    def _read_context_definition(self,
            context: Dict[str, Union[str, Dict]],
            base_url: str,
            remote_contexts: Set[str],
            override_protected: bool, validate_scoped: bool):
        # 5.5)
        version: object = context.get(VERSION)
        if version is not None:
            if self._processing_mode == JSONLD10:
                raise ProcessingModeConflictError
            if isinstance(version, float) and version == 1.1:
                self._version = version
            else:
                raise InvalidVersionValueError

        # 5.6)
        if IMPORT in context:
            context = self._handle_import(context, base_url)

        # 5.7)
        if BASE in context and len(remote_contexts) == 0:
            # 5.7.1)
            base: object = context[BASE]
            # 5.7.2)
            if base is None:
                # TODO: spec problem; specify whether base_iri can be null
                self.base_iri = "" # None
            # 5.7.3)
            # 5.7.4)
            elif self.base_iri is not None and isinstance(base, str):
                self.base_iri = resolve_iri(self.base_iri, base)
            elif isinstance(base, str) and is_iri(base):
                self.base_iri = base
            # 5.7.5)
            else:
                raise InvalidBaseIriError

        # 5.8)
        if VOCAB in context:
            # 5.8.1)
            vocab: object = context[VOCAB]
            # 5.8.2)
            if vocab is None:
                self.vocabulary_mapping = None
            # 5.8.3)
            elif isinstance(vocab, str) and (is_iri_ref(vocab) or is_blank(vocab)):
                self.vocabulary_mapping = self.expand_doc_relative_vocab_iri(vocab)
                # NOTE: The use of blank node identifiers to value for @vocab is
                # obsolete, and may be removed in a future version of JSON-LD.
            else:
                raise InvalidVocabMappingError

        # 5.9)
        if LANGUAGE in context:
            lang: object = context[LANGUAGE]
            if lang is None:
                self.default_language = None
            elif isinstance(lang, str):
                if not is_lang_tag(lang):
                    warning(f'Language tag {lang} in context is not well-formed')
                self.default_language = lang.lower()
            else:
                raise InvalidDefaultLanguageError

        # 5.10)
        if DIRECTION in context:
            direction: object = context[DIRECTION]
            if direction is None or isinstance(direction, str) and direction in DIRECTIONS:
                self.default_base_direction = direction
            else:
                raise InvalidBaseDirectionError(str(direction))

        # 5.11)
        if PROPAGATE in context:
            propagate: object = context[PROPAGATE]
            # 5.11.1)
            if self._processing_mode == JSONLD10:
                raise InvalidContextEntryError
            # 5.11.2)
            if not isinstance(propagate, bool):
                raise InvalidPropagateValueError(str(propagate))
            # NOTE: propagate is set above on the condition that local_context
            # is not an array

        # 5.12)
        defined: Dict[str, bool] = {}

        # 5.13)
        for (key, value) in context.items():
            if key in CONTEXT_KEYWORDS:
                continue
            # invoke the Create Term Definition algorithm, passing
            # result for active context,
            # context for local context,
            # key, defined, base URL,
            # TODO: pass these too:
            # the value of the @protected entry from context, if any, for protected,
            # override protected,
            # and a copy of remote contexts.
            #self.terms[key] = # TODO: set in Term (move that up here if obscure...)
            isprotected: bool = cast(bool, context.get(PROTECTED))
            Term(self, context, key, value, defined, base_url, isprotected, override_protected)

    def _handle_import(self, context: Dict[str, Union[str, Dict]], base_url: str) -> Dict:
        import_value: object = context[IMPORT]
        # 5.6.1)
        if self._processing_mode == JSONLD10:
            raise InvalidContextEntryError
        # 5.6.2)
        if not isinstance(import_value, str):
            raise InvalidImportValueError(str(import_value))
        # 5.6.3)
        import_value = resolve_iri(base_url, import_value)
        # 5.6.4)
        context_document: object = self._load_document(import_value)
        # 5.6.5)
        # TODO: LoadingDocumentFailedError
        # 5.6.6)
        if not isinstance(context_document, Dict) or CONTEXT not in context_document:
            raise InvalidRemoteContextError
        import_context: object = context_document[CONTEXT]
        if not isinstance(import_context, Dict):
            raise InvalidRemoteContextError
        # 5.6.7)
        if IMPORT in import_context:
            raise InvalidContextEntryError
        # 5.6.8)
        import_context.update(context)
        del import_context[IMPORT]
        return import_context

    def expand_vocab_iri(self, value: str) -> Optional[str]:
        return self.expand_iri(value, None, None, False, True)

    def expand_doc_relative_iri(self, value: str) -> Optional[str]:
        return self.expand_iri(value, None, None, True, False)

    def expand_doc_relative_vocab_iri(self, value: str) -> Optional[str]:
        return self.expand_iri(value, None, None, True, True)

    def _expand_init_vocab_iri(self,
            value: str,
            local_context: Dict[str, Union[str, Dict]],
            defined: Dict[str, bool],
            ) -> Optional[str]:
        return self.expand_iri(value, local_context, defined, False, True)

    # TODO: spec errata: spec says vocab=true in topic dfn but false in algorithm defintion!
    def expand_iri(self,
            value: str,
            local_context: Optional[Dict[str, Union[str, Dict]]] = None,
            defined: Optional[Dict[str, bool]] = None,
            doc_relative=False,
            vocab=False) -> Optional[str]:
        # 1)
        if value in KEYWORDS or value is None:
            return value

        # 2)
        if has_keyword_form(value):
            warning(f'Id {value} looks like a keyword')
            return None

        # 3)
        if local_context is not None and value in local_context and defined and (value not in defined or defined[value] is not True):
            Term(self, local_context, value, local_context[value], defined)

        iri_term: Optional[Term] = self.terms.get(value)

        # 4)
        if iri_term and iri_term.iri in KEYWORDS:
            return iri_term.iri

        # 5)
        if vocab and iri_term:
            return iri_term.iri

        # 6)
        if len(value) > 1 and ':' in value[1:]:
            # 6.1)
            idx: int = value.index(':')
            prefix: str = value[0:idx]
            suffix: str = value[idx + 1:]
            #prefix, suffix = value.split(':', 1)
            # 6.2)
            if prefix == '_' or suffix.startswith('//'):
                return value

            # 6.3)
            # TODO: spec problem; see step 3 above
            if local_context is not None and prefix in local_context and defined:
                if prefix not in defined or defined[prefix] is not True:
                    Term(self, local_context, prefix, local_context[prefix], defined)

            # 6.4)
            pfx_term: Optional[Term] = self.terms.get(prefix)
            if pfx_term and pfx_term.iri is not None and pfx_term.is_prefix:
                return pfx_term.iri + suffix

            # 6.5)
            # TODO: spec problem; is frag-id not an IRI here? (See TC 109)
            if not value.startswith('#') and is_iri(value):
                return value

        # 7)
        if vocab and self.vocabulary_mapping:
            return self.vocabulary_mapping + value

        # 8)
        elif doc_relative:
            return resolve_iri(self.base_iri, value)

        # 9)
        return value


class Term:

    iri: str # IRI
    is_prefix: bool
    is_protected: bool
    is_reverse_property: bool
    base_url: Optional[str] # IRI # TODO: is this proper and good?
    has_local_context: bool
    container: List[str] # empty means "optional" here...
    direction: Optional[str] # Direction
    index: Optional[str]
    language: Optional[str] # Language
    nest_value: Optional[str]
    type_mapping: Optional[str] # IRI

    _local_context: Optional[object]
    _cached_contexts: Dict[str, Optional[Context]]
    _remote_contexts: Set

    def __init__(self,
        active_context: Context,
        local_context: Dict[str, Union[str, Dict]],
        term: str,
        value: Union[str, Dict[str, object]],
        defined: Dict[str, bool],
        base_url: str = None,
        isprotected=False, # NOTE: 'protected' in spec; but a common reserved keyword
        override_protected=False, # which is used to allow changes to protected terms
        remote_contexts: Set = None, # which is used to detect cyclical context inclusions
        validate_scoped=True # which is used to limit recursion when validating possibly recursive scoped contexts
        ):
        # TODO: check redundancies of these and conditional initialization below
        self.is_prefix = False
        self.is_protected = isprotected if isinstance(isprotected, bool) else False
        self.is_reverse_property = False
        self.container = []
        self.direction = None
        self.index = None
        self.language = None
        self.nest_value = None
        self.type_mapping = None

        if remote_contexts is None:
            remote_contexts = set()

        self.base_url = None
        self.has_local_context = False
        self._local_context = None
        self._remote_contexts = remote_contexts
        self._cached_contexts = {}

        # 1)
        # TODO: place this step outside, in the call to Term?
        if term in defined:
            defined_term: bool = defined[term]
            if defined_term:
                return
            else:
                raise CyclicIriMappingError(term)
        # 2)
        if term == "":
            raise InvalidTermDefinitionError

        defined[term] = False

        # 3)
        #value = local_context[term]

        # 4)
        if term == TYPE:
            if active_context._processing_mode == JSONLD10:
                raise KeywordRedefinitionError
            # TODO: with only either or both
            if not isinstance(value, Dict) or not (
                    value.get(CONTAINER) == SET or PROTECTED in value):
                raise KeywordRedefinitionError

        # 5)
        elif term in KEYWORDS:
            raise KeywordRedefinitionError(term)

        if has_keyword_form(term):
            warning(f'Term {term} looks like a keyword'
                    ' (it matches the ABNF rule "@"1*ALPHA from [RFC5234])')

        # 6)
        prev_dfn: Optional[Term] = active_context.terms.pop(term, None)

        simple_term: bool

        # 7) + 8)
        dfn: Dict
        if value is None or isinstance(value, str):
            dfn = {}
            dfn[ID] = value
            simple_term = isinstance(value, str)
        # 9)
        else:
            if not isinstance(value, Dict):
                raise InvalidTermDefinitionError(str(value))
            dfn = value
            simple_term = False

        # 10)
        # TODO: we're doing everything *inside* of Term; move out to function + call Term?

        # 11)
        if PROTECTED in dfn:
            if active_context._processing_mode == JSONLD10:
                raise InvalidTermDefinitionError(str(dfn))
            is_protected: object = dfn[PROTECTED]
            if isinstance(is_protected, bool):
                self.is_protected = is_protected
            else:
                raise InvalidProtectedValueError

        # 12)
        type_mapping: object = dfn.get(TYPE)
        # 12.1)
        if isinstance(type_mapping, str):
            # 12.2 + 12.5)
            self.type_mapping = active_context._expand_init_vocab_iri(type_mapping, local_context, defined)
            # 12.3)
            if active_context._processing_mode == JSONLD10 and \
                    self.type_mapping in {JSON, NONE}:
                raise InvalidTypeMappingError
            # 12.4)
            elif self.type_mapping not in {ID, JSON, NONE, VOCAB} and not is_iri(self.type_mapping):
                raise InvalidTypeMappingError(self.type_mapping)
        elif type_mapping:
            raise InvalidTypeMappingError

        # 13)
        if REVERSE in dfn:
            rev: object = dfn[REVERSE]
            # 13.1)
            if ID in dfn or NEST in dfn:
                raise InvalidReversePropertyError
            # 13.2)
            if not isinstance(rev, str):
                raise InvalidIriMappingError

            # 13.3)
            if has_keyword_form(rev):
                warning(f'Reverse {rev} for term {term} has the form of a keyword')
                return
            # 13.4)
            self.iri = cast(str, active_context._expand_init_vocab_iri(rev, local_context, defined))
            if not (is_blank(self.iri) or is_iri(self.iri)):
                raise InvalidIriMappingError
            # 13.5)
            if CONTAINER in dfn:
                self.container = sorted(as_list(dfn[CONTAINER]))
                if dfn[CONTAINER] not in {SET, INDEX, None}:
                    raise InvalidReversePropertyError
            # 13.6)
            self.is_reverse_property = True
            # 13.7)
            active_context.terms[term] = self
            defined[term] = True
            return

        # 14)
        if ID in dfn and term != dfn[ID]:
            id: str = cast(str, dfn[ID]) # TODO: redundant cast for transpile
            # 14.1
            if id is None:
                # TODO: spec says "is retained to detect future redefinitions"; like this?
                self.iri = None
            # 14.2)
            else:
                # 14.2.1)
                if not isinstance(id, str):
                    raise InvalidIriMappingError
                # 14.2.2)
                if id not in KEYWORDS and has_keyword_form(id):
                    warning(f'Id {id} for term {term} is not a keyword but'
                            ' has the form of a keyword')
                # 14.2.3)
                self.iri = cast(str, active_context._expand_init_vocab_iri(id, local_context, defined))
                if self.iri is None or not (self.iri in KEYWORDS or is_blank(self.iri) or is_iri(self.iri)):
                    # TODO: check if these are to be skipped instead of halting
                    return
                    #raise InvalidIriMappingError(f'{term}: {self.iri}')
                if self.iri == CONTEXT:
                    raise InvalidKeywordAliasError

                # 14.2.4)
                if ':' in term[1:-1] or '/' in term:
                    # 14.2.4.1)
                    defined[term] = True
                    # 14.2.4.2)
                    if active_context._expand_init_vocab_iri(term, local_context, defined) != self.iri:
                        # TODO: spec or TC expand 0026 problem; just pass works
                        # TODO: expansion TC 0071 fails unless we just return
                        pass#return
                        #raise InvalidIriMappingError(term, self.iri)
                # 14.2.5) # TODO: really no ':' anywhere, so elif not strictly correct...
                elif simple_term and (self.iri[-1] in PREFIX_DELIMS or is_blank(self.iri)):
                    self.is_prefix = True

        # 15)
        elif ':' in term[1:]:
            # 15.1)
            idx: int = term.index(':')
            prefix: str = term[0:idx]
            suffix: str = term[idx + 1:]
            #prefix, suffix = term.split(':', 1)

            if prefix in local_context:
                Term(active_context, local_context, prefix, local_context[prefix], defined)

            # 15.2)
            if prefix in active_context.terms:
                self.iri = active_context.terms[prefix].iri + suffix
            # 15.3)
            else:
                #assert is_iri(term) or is_blank(term)
                self.iri = term

        # 16)
        elif '/' in term:
            # 16.1) assert is_relative_iri(term)
            # 16.2)
            self.iri = cast(str, active_context.expand_vocab_iri(term))
            if not is_iri(self.iri):
                raise InvalidIriMappingError

        # 17)
        elif term == TYPE:
            self.iri = TYPE

        # 18)
        elif active_context.vocabulary_mapping:
            self.iri = active_context.vocabulary_mapping + term
        else:
            raise InvalidIriMappingError(f'{term}: {str(value)}')

        # 19)
        if CONTAINER in dfn:
            container: object = dfn[CONTAINER]
            # 19.1)
            container_terms: Optional[Set] = None
            if isinstance(container, List):
                container_terms = set(container) # TODO: just - {SET} ?
                if SET in container_terms and LIST not in container_terms:
                    container_terms.remove(SET)
            if not (isinstance(container, str) and
                container in CONTAINER_KEYWORDS
                ) and not (
                    container_terms is not None and (
                        len(container_terms) == 0
                        or len(container_terms) == 1 and all(t in CONTAINER_KEYWORDS for t in container_terms)
                        or container_terms == {GRAPH, ID}
                        or container_terms == {GRAPH, INDEX}
                        or list(container_terms)[0] in {INDEX, GRAPH, ID, TYPE, LANGUAGE}
                    )
                ):
                raise InvalidContainerMappingError(container)

            # 19.2)
            if active_context._processing_mode == JSONLD10:
                if not isinstance(container, str) or container in {GRAPH, ID, TYPE}:
                    raise InvalidContainerMappingError(str(container))

            # 19.3)
            self.container = sorted(as_list(cast(object, container)))

            # 19.4)
            if TYPE in self.container:
                # 19.4.1)
                if self.type_mapping is None:
                    self.type_mapping = ID
                # 19.4.2)
                elif self.type_mapping not in {ID, VOCAB}:
                    raise InvalidTypeMappingError

        # 20)
        if INDEX in dfn:
            # 20.1)
            if active_context._processing_mode == JSONLD10 or INDEX not in self.container:
                raise InvalidTermDefinitionError(str(value))
            # 20.2)
            index: object = dfn[INDEX]
            if not isinstance(index, str):
                raise InvalidTermDefinitionError(str(value))
            if not is_iri(active_context.expand_vocab_iri(index)):
                raise InvalidTermDefinitionError(str(value))
            # 20.3)
            self.index = index

        # 21)
        self.has_local_context = CONTEXT in dfn
        if self.has_local_context:
            # 21.1)
            if active_context._processing_mode == JSONLD10:
                raise InvalidTermDefinitionError(str(dfn))
            # 21.2)
            # 21.3)
            # TODO: doesn't produce the expected effects (see c032, c033)
            #try:
            #    self.get_local_context(active_context)
            #except JsonLdError as e:
            #    raise InvalidScopedContextError(term)
            # 21.4)
            # TODO: different from spec; since propagate behaviour may be based
            # on usage as @type (see 5f496df9) we use a memoized approach.
            self._local_context = dfn[CONTEXT]
            self.base_url = base_url

        # 22)
        if LANGUAGE in dfn and TYPE not in dfn:
            # 22.1)
            lang: object = dfn[LANGUAGE]
            if not isinstance(lang, str) and lang is not None:
                raise InvalidLanguageMappingError
            if not is_lang_tag(lang):
                warning(f'Language tag {lang} in term {term} is not well-formed')
            # 22.2)
            # TODO: [5f6117d4] spec uses maps with key missing != null; note this difference
            self.language = NULL if lang is None else lang.lower()

        # 23)
        if DIRECTION in dfn and TYPE not in dfn:
            # 23.1)
            dir: object = dfn[DIRECTION]
            if dir not in DIRECTIONS and dir is not None:
                raise InvalidBaseDirectionError
            # 23.2)
            # TODO: see 5f6117d4
            self.direction = NULL if dir is None else cast(str, dir)

        # 24)
        if NEST in dfn:
            # 24.1)
            if active_context._processing_mode == JSONLD10:
                raise InvalidTermDefinitionError
            # 24.2)
            self.nest_value = cast(str, dfn[NEST]) # TODO: redundant cast for transpile
            if not isinstance(self.nest_value, str) or \
                    (self.nest_value != NEST and self.nest_value in KEYWORDS):
                raise InvalidNestValueError

        # 25)
        if PREFIX in dfn:
            # 25.1)
            if active_context._processing_mode == JSONLD10 or ':' in term or '/' in term:
                raise InvalidTermDefinitionError
            # 25.2)
            self.is_prefix = cast(bool, dfn[PREFIX]) # TODO: redundant cast for transpile
            if not isinstance(self.is_prefix, bool):
                raise InvalidPrefixValueError
            # 25.3)
            if self.is_prefix and self.iri in KEYWORDS:
                raise InvalidTermDefinitionError

        # 26) TODO:
        #if len(set(dfn.keys()) - TERM_KEYWORDS) > 0:
        #    raise InvalidTermDefinitionError

        # 27)
        if not override_protected and prev_dfn and prev_dfn.is_protected:
            # 27.2)
            self.is_protected = prev_dfn.is_protected
            # 27.1)
            if not self.matches(prev_dfn):
                raise ProtectedTermRedefinitionError 

        # 28)
        active_context.terms[term] = self
        defined[term] = True

    def get_local_context(self, active_context: Context, propagate=True) -> Context:
        cache_key: str = f'{id(active_context)}:{str(propagate)}'
        cached: Optional[Context] = self._cached_contexts.get(cache_key)

        # TODO: should be passed explicitly, but seems to correlate
        # (might even be named "type-scoped"?)
        override_protected: bool = propagate

        if cached is None:
            cached = active_context.get_context(self._local_context,
                        self.base_url,
                        set(self._remote_contexts),
                        override_protected=override_protected,
                        validate_scoped=False)
            self._cached_contexts[cache_key] = cached

        if (not isinstance(self._local_context, Dict)
            or PROPAGATE not in self._local_context):
                cached._propagate = propagate

        return cached

    def matches(self, other: object) -> bool:
        if not isinstance(other, Term):
            return False

        return self.iri == other.iri and \
                self.is_prefix == other.is_prefix and \
                self.is_reverse_property == other.is_reverse_property and \
                self.base_url == other.base_url and \
                self.has_local_context == other.has_local_context and \
                self.container == other.container and \
                self.direction == other.direction and \
                self.index == other.index and \
                self.language == other.language and \
                self.nest_value == other.nest_value and \
                self.type_mapping == other.type_mapping and \
                self._local_context == other._local_context



import json
import sys
import uuid
from http.client import HTTPResponse
from io import StringIO, TextIOWrapper
from typing import Callable, Dict, Iterator, Optional, Set, TextIO, Union
from urllib.parse import urljoin, urlparse
from urllib.request import Request, urlopen

from .builtins import Char
from .jsonld.base import JSONLD_CONTEXT_RELATION
from .mimetypes import SUFFIX_MIME_TYPE_MAP, guess_mime_type

JSONLD_MIME_TYPE = SUFFIX_MIME_TYPE_MAP['jsonld']
JSON_MIME_TYPES: Set[str] = {JSONLD_MIME_TYPE, 'application/json'}

ACCEPT_HEADER = 'Accept'

ACCEPTS = ", ".join(
    mt + (f';q=0.{10 - i}' if i > 0 else '')
    for i, mt in enumerate(SUFFIX_MIME_TYPE_MAP.values())
)


class Input:
    document_url: Optional[str]
    content_type: Optional[str]
    context_url: Optional[str]
    profile: Optional[str]

    def __init__(
        self, source: Union[str, TextIO, None] = None, headers: Optional[Dict] = None
    ):
        stream: TextIO
        if isinstance(source, str):
            self.document_url = source
            if source.startswith(('http', 'https')):
                stream = self._open_request(source, headers)
            else:
                stream = open(_remove_file_protocol(source))
                self.content_type = guess_mime_type(source)
        elif source is None:
            stream = sys.stdin
        else:
            stream = source

        self._source = stream

    def _open_request(self, source: str, headers: Optional[Dict] = None) -> TextIO:
        req = Request(source)

        if headers is None or ACCEPT_HEADER not in headers:
            req.add_header(ACCEPT_HEADER, ACCEPTS)
        if headers is not None:
            for h, hv in headers.items():
                req.add_header(h, hv)

        res = urlopen(req)

        self.content_type = res.headers.get_content_type()
        self.context_url = _get_link_header(
            res, rel=JSONLD_CONTEXT_RELATION, type=JSONLD_MIME_TYPE
        )
        self.profile = _get_link_header(res, rel='profile')

        return TextIOWrapper(res)

    @property
    def document(self) -> object:
        if self.content_type in JSON_MIME_TYPES:
            return load_json(self)
        else:
            return self.read()

    def read(self) -> str:
        return self._source.read()

    def lines(self) -> Iterator[str]:
        return self._source

    def characters(self) -> Iterator[Char]:
        return (c for l in self._source for c in l)

    def close(self):
        if self._source is not sys.stdin:
            self._source.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


class Output:
    def __init__(self, dest: Optional[TextIO] = None):
        if dest is None:
            dest = StringIO()
        self._dest = dest

    def write(self, s: str):
        print(s, file=self._dest, end='')

    def writeln(self, s: str):
        print(s, file=self._dest)

    def get_captured(self):
        return self._dest

    def get_value(self):
        assert isinstance(self._dest, StringIO)
        return self._dest.getvalue()


LoadDocumentCallback = Callable[[str], Input]


document_loader: Optional[LoadDocumentCallback] = None


def set_document_loader(loader: LoadDocumentCallback):
    global document_loader
    document_loader = loader


def load_json(source: Union[str, Input]) -> object:
    stream = None

    if not isinstance(source, Input):
        assert isinstance(source, str)
        source = document_loader(source) if document_loader else Input(source)

    if source._source == sys.stdin:
        return json.load(source._source)

    with source._source as fp:
        return json.load(fp)


def parse_json(s: str) -> object:
    return json.loads(s)


def dump_json(o: object, pretty=False) -> str:
    return json.dumps(o, indent=2 if pretty else None, ensure_ascii=not pretty)


def dump_canonical_json(o: object) -> str:
    return json.dumps(o, indent=None, separators=(',', ':'), sort_keys=True)


def resolve_iri(base: str, relative: str) -> str:
    if '//' in relative:
        url = urlparse(relative)
        urlpath: str = url.path
        if url.params:
            urlpath += f';{url.params}'
        if url.query:
            urlpath += f'?{url.query}'
        return urljoin(base, urljoin(relative, urlpath))

    return urljoin(base, relative)


def uuid4() -> str:
    return str(uuid.uuid4())


def warning(msg: str):
    print(msg, file=sys.stderr)


def _remove_file_protocol(ref: str):
    if ref.startswith('file://'):
        return ref[7:]
    elif ref.startswith('file:'):
        return ref[5:]
    return ref


def _get_link_header(
    response: HTTPResponse, rel: str, type: Optional[str] = None
) -> Optional[str]:
    links = response.headers.get_all('Link')
    if links is None:
        return None

    for link in links:
        href = link[link.find('<') : link.find('>')]
        if rel in link and (type is None or type in link):
            return href

    return None



import argparse
import json
import sys
from typing import Any

from .common import Input, Output, dump_json, load_json
from .jsonld.compaction import compact
from .jsonld.expansion import expand
from .jsonld.flattening import flatten
from .mimetypes import SUFFIX_MIME_TYPE_MAP


def eprint(msg):
    print(msg, file=sys.stderr)


def parse_rdf(source, fmt):
    headers = {}
    accept = SUFFIX_MIME_TYPE_MAP.get(fmt)
    if accept:
        headers['Accept'] = accept

    inp = Input(None if source == '-' else source, headers)

    # TODO: Use inp.content_type (which uses guess_mime_type)?
    # So instead of using suffixes or "names" below, check mime-type (accept).
    if fmt is None and isinstance(source, str):
        fmt = source.rsplit('.', 1)[-1]

    if fmt in {'trig', 'ttl', 'turtle'}:
        from .trig import parser as trig
        return trig.parse(inp)

    if fmt in {'nq', 'nt'}:
        from .nq import parser as nq
        return nq.parse(inp)

    return load_json(inp)


def serialize_rdf(result, fmt):
    if fmt is None or fmt == 'jsonld':
        print(dump_json(result, pretty=True))
        return

    out = Output(sys.stdout)
    if fmt in {'trig', 'ttl', 'turtle', 'turtle-union'}:
        from .trig import serializer as trig
        if fmt == 'trig':
            trig.serialize(result, out)
        else:
            trig.serialize_turtle(result, out, union=fmt == 'turtle-union')
    else:
        from .jsonld.rdf import to_rdf_dataset
        dataset = to_rdf_dataset(result)
        if fmt == 'nq':
            from .nq import serializer as nq
            nq.serialize(dataset, out)
        else:
            print(json.dumps(dataset, indent=2,
                             default=lambda o: o.__dict__))


def process_source(source, context_data, args, ordered=True):
    base_iri = (
        args.base if args.base
        else f'file:///dev/stdin' if source == '-'
        else f'file://{source}'
    )

    try:
        data = parse_rdf(source, args.input)

        if args.expand_context:
            expand_context = (
                None if args.expand_context == True else args.expand_context
            )
            result: Any = expand(data, base_iri, expand_context, ordered=ordered)
        else:
            result = data

        if args.flatten or args.output in {'nq', True}:
            result = flatten(result, ordered=ordered)
        if context_data:
            result = compact(context_data, result, base_iri, ordered=ordered)

        serialize_rdf(result, args.output)

    except Exception as e:
        eprint(f"Error in file '{source}'")
        import traceback
        traceback.print_exc()


def main():
    argparser = argparse.ArgumentParser()
    argparser.add_argument('source', metavar='SOURCE', nargs='*')
    argparser.add_argument('-c', '--context', help='Use to compact expanded JSON-LD')
    argparser.add_argument('-e', '--expand-context', const=True, nargs='?',
                        help='Use to expand plain JSON to JSON-LD')
    argparser.add_argument('-b', '--base',
                        help='Set the base IRI (default is current source)')
    argparser.add_argument('-f', '--flatten', action='store_true')
    argparser.add_argument('-i', '--input-format', help='Set RDF input format')
    argparser.add_argument('-o', '--output-format', help='Set RDF output format')

    args = argparser.parse_args()

    context_data = None
    if args.context:
        context_data = load_json(args.context)

    sources = args.source or ['-']

    for source in sources:
        if len(sources) > 1:
            eprint(f"Parsing file: '{source}'")

        process_source(source, context_data, args)


if __name__ == '__main__':
    main()



from typing import Optional, Dict, List, Set, Union, cast


Scalar = Union[str, int, float, bool]
JsonObject = Union[Dict, List, Scalar]
JsonMap = Dict[str, JsonObject]
JsonOptMap = Dict[str, Optional[JsonObject]]
JsonList = List[JsonObject]
JsonOptList = List[Optional[JsonObject]]


# From <https://tools.ietf.org/html/rfc3986#section-2.2>
PREFIX_DELIMS: Set[str] = {':', '/', '?', '#', '[', ']', '@'}

BASE: str = '@base'
CONTAINER: str = '@container'
CONTEXT: str = '@context'
DIRECTION: str = '@direction'
GRAPH: str = '@graph'
ID: str = '@id'
IMPORT: str = '@import'
INCLUDED: str = '@included'
INDEX: str = '@index'
JSON: str = '@json'
LANGUAGE: str = '@language'
LIST: str = '@list'
NEST: str = '@nest'
NONE: str = '@none'
PREFIX: str = '@prefix'
PROPAGATE: str = '@propagate'
PROTECTED: str = '@protected'
REVERSE: str = '@reverse'
SET: str = '@set'
TYPE: str = '@type'
VALUE: str = '@value'
VERSION: str = '@version'
VOCAB: str = '@vocab'

# TODO: not "really" (public) keywords...?
ANY: str = '@any'
NULL: str = '@null'
DEFAULT: str = '@default'

# TODO: see 5f6117d4
NULLS: Set[Optional[str]] = {None, NULL}
NOTHING: Set[Optional[str]] = {None, NULL, NONE}

KEYWORDS: Set[str] = {
    BASE,
    CONTAINER,
    CONTEXT,
    DIRECTION,
    GRAPH,
    ID,
    IMPORT,
    INCLUDED,
    INDEX,
    JSON,
    LANGUAGE,
    LIST,
    NEST,
    NONE,
    PREFIX,
    PROPAGATE,
    PROTECTED,
    REVERSE,
    SET,
    TYPE,
    VALUE,
    VERSION,
    VOCAB
}

CONTEXT_KEYWORDS: Set[str] = {
    VERSION,
    IMPORT,
    BASE,
    VOCAB,
    LANGUAGE,
    DIRECTION,
    PROPAGATE,
    PROTECTED
}

VALUE_KEYWORDS: Set[str] = {
    DIRECTION, INDEX, LANGUAGE, TYPE, VALUE
}

CONTAINER_KEYWORDS: Set[str] = {GRAPH, ID, INDEX, LANGUAGE, LIST, SET, TYPE}

DIRECTIONS: Set[str] = {'rtl', 'ltr'}

JSONLD10: str = 'json-ld-1.0'
JSONLD11: str = 'json-ld-1.1'

JSONLD_CONTEXT_RELATION: str = 'http://www.w3.org/ns/json-ld#context'


class JsonLdError(Exception):
    pass


def is_iri(value: Optional[str]) -> bool:
    return value is not None and ':' in value and is_iri_ref(value)


def is_iri_ref(value: Optional[str]) -> bool:
    ... # TODO: check legal chars
    return value is not None and ' ' not in value and not is_blank(value)


def is_blank(value: str) -> bool:
    ...
    return value.startswith('_:')


def has_keyword_form(s: str) -> bool:
    return s.startswith('@') and s[1:].isalpha()


def is_lang_tag(value: Optional[str]) -> bool:
    ... # TODO: according to section 2.2.9 of [BCP47]
    return value is not None and value[0].isalpha() and  ' ' not in value


# TODO: spec errata: spec def excludes null but links to whatwg which includes it
def is_scalar(o: object) -> bool:
    return isinstance(o, (str, int, float, bool))


def is_graph_object(o: JsonMap) -> bool:
    if GRAPH in o:
        if ID in o:
            if INDEX in o:
                return len(o) == 3
            return len(o) == 2
        if INDEX in o:
            return len(o) == 2
        return len(o) == 1
    return False


def is_simple_graph_object(o: JsonMap) -> bool:
    if GRAPH in o:
        return len(o) == 2 if INDEX in o else len(o) == 1
    return False


def add_value_as_list(map: Dict, key: str, value: object):
    add_value(map, key, value, True)


def add_value(map: Dict, key: str, value: object, aslist=False):
    # TODO: rework this according to spec..
    existing: object = map.get(key)
    if aslist and key not in map:
        map[key] = as_list(value)
    #elif isinstance(existing, List):
    #    map.pop(key)
    #    for v in existing:
    #        add_value(map, key, v, aslist)
    elif key not in map:
        map[key] = value
    else:
        if not isinstance(existing, List):
            map[key] = [existing]
        if isinstance(value, List):
            map[key] += value
        else:
            assert isinstance(map, Dict) # TODO: just for transpile
            values: object = map[key]
            assert isinstance(values, List) # TODO: just for transpile
            assert isinstance(value, object) # TODO: just for transpile
            values.append(value)


def as_list(obj: object) -> List:
    return obj if isinstance(obj, List) else [obj]


def relativise_iri(base: str, iri: str) -> str:
    if iri.startswith(base + '#'):
        return iri[len(base):]
    if '?' in iri and iri.startswith(base):
        return iri[len(base):]
    if not base.endswith('/'):
        last: int = base.rfind('/')
        base = base[0:last + 1]
    if iri.startswith(base):
        return iri[len(base):]

    parentbase: str = base[:base.rfind('/')]
    leaf: str = iri[iri.rfind('/') + 1:]
    relativeto: List[str] = []
    while '/' in parentbase and not parentbase.endswith(':/'):
        if iri.startswith(parentbase):
            relativeto.append(leaf)
            return '/'.join(relativeto)
        relativeto.append('..')
        parentbase = parentbase[:parentbase.rfind('/')]

    return iri


def node_equals(a: JsonObject, b: JsonObject) -> bool:
    if is_scalar(a):
        return type(a) == type(b) and a == b
    if isinstance(a, List):
        if not isinstance(b, List):
            return False
        # TODO: transpile or live with C-style code...
        #return all(node_equals(it, b[i]) for i, it in enumerate(a))
        i: int = 0
        for ai in a:
            if not node_equals(ai, b[i]):
                return False
            i += 1
        return True
    elif isinstance(a, Dict):
        if not isinstance(b, Dict):
            return False
        return all(k in b and node_equals(a[k], b[k]) for k in a.keys())
    else:
        return False
